@@ -24,13 +24,12 @@
 
 #define mb() do {  asm volatile(__ASM_BARRIER : : : "memory"); } while (0)
 
-static inline void osb(void)
-{
-	asm volatile(
-		ALTERNATIVE("", ".long 0xb2e8f000", 81)
-		: : : "memory");
-}
-#define osb osb
+/* Prevent speculative execution past this barrier. */
+#define barrier_nospec()						\
+	do {								\
+		asm volatile(ALTERNATIVE("", ".long 0xb2e8f000", 81)	\
+			     : : : "memory");				\
+	} while (0)
 
 #define rmb()				barrier()
 #define wmb()				barrier()
@@ -63,4 +62,28 @@ do {									\
 	___p1;								\
 })
 
+/**
+ * array_index_mask_nospec - generate a mask for array_idx() that is
+ * ~0UL when the bounds check succeeds and 0 otherwise
+ * @index: array element index
+ * @size: number of elements in array
+ */
+#define array_index_mask_nospec array_index_mask_nospec
+static inline unsigned long array_index_mask_nospec(unsigned long index,
+						    unsigned long size)
+{
+	unsigned long mask;
+
+	if (__builtin_constant_p(size) && size > 0) {
+		asm("	clgr	%2,%1\n"
+		    "	slbgr	%0,%0\n"
+		    :"=d" (mask) : "d" (size-1), "d" (index) :"cc");
+		return mask;
+	}
+	asm("	clgr	%1,%2\n"
+	    "	slbgr	%0,%0\n"
+	    :"=d" (mask) : "d" (size), "d" (index) :"cc");
+	return ~mask;
+}
+
 #endif /* __ASM_BARRIER_H */
