@@ -1376,7 +1376,7 @@ module_param_named(logsyscall, aa_g_logsyscall, aabool, S_IRUGO | S_IWUSR);
 
 /* Maximum pathname length before accesses will start getting rejected */
 unsigned int aa_g_path_max = 2 * PATH_MAX;
-module_param_named(path_max, aa_g_path_max, aauint, S_IRUGO | S_IWUSR);
+module_param_named(path_max, aa_g_path_max, aauint, S_IRUGO);
 
 /* Determines how paranoid loading of policy is and how much verification
  * on the loaded policy is done.
